import argparse
from pathlib import Path
import shutil

from colorama import init, Fore, Style
from gasflux.processing_pipelines import process_main

init()


def find_config_file(path: Path, recursive: bool = True):
    if recursive:
        config_files = [file for file in path.rglob("*.yaml") if "gasflux_config" in file.read_text()]
    else:
        config_files = [file for file in path.glob("*.yaml") if "gasflux_config" in file.read_text()]
    if len(config_files) == 1:
        return config_files[0]
    elif len(config_files) > 1:
        raise ValueError(
            "Multiple candidate config files found: {}".format(", ".join(str(file) for file in config_files))
        )
    else:
        raise FileNotFoundError("No config file found in the supplied path or its child folders")


def process_command(data_path: str, config_path: str, test: bool):
    if test:
        data_file = Path(__file__).parent / "testdata" / "testdata.csv"
        config_file = Path(__file__).parent / "testdata" / "testconfig.yaml"
        process_main(data_file, config_file)
        return

    dpath_obj = Path(data_path)

    if dpath_obj.is_dir():
        data_files = list(dpath_obj.rglob("*.csv"))
        if not data_files:
            raise FileNotFoundError(f"No CSV files found in directory: {data_path}")
    elif dpath_obj.is_file():
        data_files = [dpath_obj]
    else:
        raise FileNotFoundError(f"Invalid data path: {data_path}")

    if config_path is None:
        try:
            config_file = find_config_file(dpath_obj if dpath_obj.is_dir() else dpath_obj.parent)
        except FileNotFoundError:
            response = (
                input(
                    "No configuration file found. Generate config file? \n"
                    "You may have to modify it before processing. [y/n]: "
                )
                .strip()
                .lower()
            )

            if response == "y":
                config_file = (dpath_obj if dpath_obj.is_dir() else dpath_obj.parent) / "gasflux_config.yaml"
                shutil.copy(Path(__file__).parent / "gasflux_config.yaml", config_file)
                print(f"Config file copied to: {config_file}")
            else:
                print("No config file generated. Exiting.")
                return
    else:
        config_file = Path(config_path)

    for data_file in data_files:
        process_main(data_file, config_file)


def generate_config_command(config_destination: str, recursive: bool = False, template_path: str | None = None):
    if template_path:
        template = Path(template_path)
        print(f"Using custom template: {template}")
    else:
        template = Path(__file__).parent / "gasflux_config.yaml"

    if not template.exists():
        raise FileNotFoundError(f"Template file not found: {template}")

    destination_path = Path(config_destination)

    if not destination_path.is_dir():
        raise NotADirectoryError(f"Destination path is not a directory: {config_destination}")

    overwrite_all = False  # Flag to determine if we should overwrite all files without prompting

    if recursive:
        for subdir in destination_path.rglob("*"):
            if subdir.is_dir():
                config_file = subdir / "gasflux_config.yaml"
                if config_file.exists() and not overwrite_all:
                    response = (
                        input(f"Config file already exists at {config_file}. Overwrite? [y/n/a/c]: ").strip().lower()
                    )
                    if response == "a":
                        overwrite_all = True
                    elif response == "c":
                        print("Operation canceled.")
                        return
                    elif response == "n":
                        print(f"Skipping {config_file}")
                        continue
                shutil.copy(template, config_file)
                print(f"Config file copied to: {config_file}")
    else:
        config_file = destination_path / "gasflux_config.yaml"
        if config_file.exists() and not overwrite_all:
            response = input(f"Config file already exists at {config_file}. Overwrite? [y/n/a/c]: ").strip().lower()
            if response == "a":
                overwrite_all = True
            elif response == "c":
                print("Operation canceled.")
                return
            elif response == "n":
                print(f"Skipping {config_file}")
                return
        shutil.copy(template, config_file)
        print(f"Config file copied to: {config_file}")


def main_cli():
    parser = argparse.ArgumentParser(
        description="GasFlux Processing Pipeline",
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,  # Disable the default help to use the custom help screen
    )
    subparsers = parser.add_subparsers(dest="command")

    process_parser = subparsers.add_parser(
        "process",
        help="Process CSV data files",
        description=(
            "Process CSV data files located in a specified directory or a single file.\n"
            "The program will search for a configuration file (gasflux_config.yaml) in the same directory as the data file(s).\n"  # noqa
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    process_parser.add_argument(
        "data_path",
        nargs="?",
        default=None,
        help=(
            "Path to the data file or directory containing CSV files.\n"
            "If a directory is provided, all CSV files within the directory will be processed."
        ),
    )
    process_parser.add_argument(
        "--config-path",
        "-c",
        default=None,
        help=(
            "Path to the configuration file (gasflux_config.yaml).\n"
            "If not specified, the program will search for the configuration file in the same directory as the data file(s).\n"  # noqa
            "If no configuration file is found, the default configuration will be used."
        ),
    )
    process_parser.add_argument(
        "--test",
        action="store_true",
        help="Use test data instead of the specified data file(s).",
    )

    config_parser = subparsers.add_parser(
        "generate-config",
        help="Generate a configuration file",
        description=(
            "Generate a configuration file (gasflux_config.yaml) in the specified directory.\n"
            "If no directory is specified, the configuration file will be generated in the current directory."
            "Use flag --recursive to generate config files in all subdirectories recursively. YOu might want to"
            "do this if you want to have a separate config for each data file"
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    config_parser.add_argument(
        "--recursive",
        action="store_true",
        help="Generate config files in all subdirectories recursively.",
    )

    config_parser.add_argument(
        "--template",
        default=None,
        help="Path to a custom template configuration file.",
    )

    config_parser.add_argument(
        "config_destination",
        nargs="?",
        default=None,
        help="Destination directory for the generated configuration file.",
    )

    # Custom help option
    parser.add_argument(
        "-h",
        "--help",
        action="store_true",
        help="Show this help message and exit",
    )

    args = parser.parse_args()

    if args.command == "process":
        process_command(args.data_path, args.config_path, args.test)
    elif args.command == "generate-config":
        generate_config_command(args.config_destination, args.recursive, args.template)
    elif args.help:
        display_help()
    else:
        display_help()


def display_help():
    help_text = f"""
{Fore.CYAN}GasFlux Processing Pipeline{Style.RESET_ALL}
{Fore.CYAN}==========================={Style.RESET_ALL}

{Fore.YELLOW}Description:{Style.RESET_ALL}
The GasFlux Processing Pipeline is a command-line tool for processing CSV data files and generating configuration files.

{Fore.YELLOW}Usage:{Style.RESET_ALL}
  gasflux [OPTIONS] COMMAND [ARGS]...

{Fore.YELLOW}Commands:{Style.RESET_ALL}
  {Fore.GREEN}process{Style.RESET_ALL}         Process CSV data files
  {Fore.GREEN}generate-config{Style.RESET_ALL} Generate a default configuration file

{Fore.YELLOW}Options:{Style.RESET_ALL}
  {Fore.BLUE}-h, --help{Style.RESET_ALL}      Show this help message and exit
  {Fore.BLUE}-v, --version{Style.RESET_ALL}   Show the version and exit

{Fore.YELLOW}Commands Help:{Style.RESET_ALL}
  {Fore.GREEN}process{Style.RESET_ALL}         Process CSV data files located in a specified directory or a single file.
                  The program will search for a singular configuration file (gasflux_config.yaml) in the same directory
                  - if a directory is supplied to [DATA_PATH] then child directories are searched too.

                  {Fore.YELLOW}Usage:{Style.RESET_ALL} gasflux process [OPTIONS] [DATA_PATH]

                  {Fore.YELLOW}Options:{Style.RESET_ALL}
                    {Fore.BLUE}-c, --config-path PATH{Style.RESET_ALL}  Path to the configuration file (gasflux_config.yaml).
                                            If not specified, the program will search for the configuration file
                                            in the same directory as the data file(s).
                                            If no configuration file is found, the default configuration will be used.
                    {Fore.BLUE}--test{Style.RESET_ALL}                  Use test data instead of the specified data file(s).
                    {Fore.BLUE}-h, --help{Style.RESET_ALL}              Show this help message and exit

  {Fore.GREEN}generate-config{Style.RESET_ALL} Generate a default configuration file (gasflux_config.yaml) in the specified directory.
                  If no directory is specified, the configuration file will be generated in the current directory.

                  {Fore.YELLOW}Usage:{Style.RESET_ALL} gasflux generate-config [OPTIONS] [CONFIG_DESTINATION]

                  {Fore.YELLOW}Options:{Style.RESET_ALL}
                    {Fore.BLUE}-h, --help{Style.RESET_ALL}  Show this help message and exit

{Fore.YELLOW}Examples:{Style.RESET_ALL}
  Process a single CSV file:
    $ gasflux process {Fore.MAGENTA}"/path/to/data.csv"{Style.RESET_ALL}

  Process all CSV files in a directory:
    $ gasflux process {Fore.MAGENTA}"/path/to/data/directory"{Style.RESET_ALL}

  Process a single CSV file with a specific configuration file:
    $ gasflux process {Fore.MAGENTA}"/path/to/data.csv"{Style.RESET_ALL} --config-path {Fore.MAGENTA}/path/to/config.yaml{Style.RESET_ALL}

  Process test data:
    $ gasflux process {Fore.BLUE}--test{Style.RESET_ALL}

  Generate a default configuration file in the current directory:
    $ gasflux generate-config {Fore.MAGENTA}.{Style.RESET_ALL}

  Generate a default configuration file in a specific directory recursively using a custom template:
    $ gasflux generate-config {Fore.MAGENTA}"/path/to/directory"{Style.RESET_ALL} {Fore.BLUE}--recursive{Style.RESET_ALL} {Fore.BLUE}--template{Style.RESET_ALL} {Fore.MAGENTA}"/path/to/template.yaml"{Style.RESET_ALL}
"""  # noqa
    print(help_text)


if __name__ == "__main__":
    main_cli()



"""functions related to gas transformations and calculations, e.g. density, point flux etc."""

import molmass
import pandas as pd

gas_variables = {
    "standard_pressure": 1013.25,  # hPa/mbar
    "standard_temperature": 273.15,  # degrees K
    "standard_molar_volume": 0.022413969545014,  # m3⋅mol-1
}


def mass(formula: str) -> float:
    """Return the molar mass of a gas in g/mol."""
    return molmass.Formula(formula.upper()).mass  # only accepts capital letters


def gas_density(local_pressure: float, local_temperature_celsius: float, gas: str) -> float:  # millibars and celsius
    """
    Calculate the density of a gas in kg/m3 based on local pressure and temperature.

    Parameters:
    - local_pressure: The local pressure in hPa/mbar.
    - local_temperature: The local temperature in degrees Celsius.
    - gas: The chemical formula of the gas.

    Returns:
    - The density of the gas in kg/m3.

    Assumes ideal gas behavior.
    """
    local_temperature_kelvin = local_temperature_celsius + gas_variables["standard_temperature"]
    local_volume = (
        gas_variables["standard_molar_volume"]
        * (gas_variables["standard_pressure"] / local_pressure)
        * ((local_temperature_kelvin + gas_variables["standard_temperature"]) / gas_variables["standard_temperature"])
    )  # m3⋅mol-1
    return mass(gas) / 1000 / local_volume


def gas_flux_column(df: pd.DataFrame, gas: str, wind: str = "windspeed") -> pd.DataFrame:
    """
    Add columns to the DataFrame for the gas density, mass, and flux.

    Parameters:
    - df: The DataFrame.
    - gas: The chemical formula of the gas.
    - wind: The column name for the wind speed (NB - must be perpendicular to the plane)

    Returns:
    - The DataFrame with the added columns.
    """
    average_temp = df["temperature"].mean()  # celsius
    average_pressure = df["pressure"].mean()  # hPa
    gd = gas_density(local_pressure=average_pressure, local_temperature_celsius=average_temp, gas=gas)  # kg/m3
    df[f"{gas}_kg_m3"] = gd * (df[f"{gas}_normalised"] * 1e-6)  # kg/m3
    df[f"{gas}_kg_h_m2"] = df[f"{gas}_kg_m3"] * df[wind] * 60 * 60  # kg/h/m2
    return df



"""Processing function, usually implying some kind of filtering or data transformation."""

from matplotlib.figure import Figure
from itertools import groupby
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import odr
from scipy.optimize import least_squares
from scipy.signal import find_peaks
from scipy.stats import circmean
import warnings
import logging


def circ_median(x: np.ndarray | pd.Series) -> float:
    """
    Finds the largest "empty" arc in the circle, splits it there and returns the median of the straight line.
    Probably there are more elegant ways... see https://github.com/scipy/scipy/issues/6644

    Parameters:
        x (np.ndarray | pd.Series): The input array of angles.

    Returns:
        float: The circular median of the input angles.
    """
    angles = sorted(x)
    n = len(angles)
    if n == 1:
        return angles[0]
    gaps = [(angles[(i + 1) % n] - angles[i]) % 360 for i in range(n)]
    largest_gap_index = gaps.index(max(gaps))
    rearranged_angles = angles[largest_gap_index + 1 :] + angles[: largest_gap_index + 1]
    if n % 2 == 1:
        return rearranged_angles[n // 2]
    else:
        return (rearranged_angles[n // 2 - 1] + rearranged_angles[n // 2]) % 360 / 2


def min_angular_displacement(x: float | np.ndarray, y: float | np.ndarray) -> float | np.ndarray:
    """
    Calculates the minimum circular difference between two angles (in 360 degree space)
    """
    return np.minimum(np.abs(x - y) % 360, (360 - np.abs(x - y)) % 360)


def wind_offset_correction(df: pd.DataFrame, plane_angle: float) -> pd.DataFrame:
    """
    Corrects wind direction data for a given plane angle, assuming the plane is the primary orientation of the dataset.
    This is useful for aligning wind data with the plane's orientation, facilitating analysis of wind effects.

    Parameters:
        df (pd.DataFrame): The input dataframe containing wind direction data.
        wind_dir_col (str): Column name for wind direction data.
        plane_angle (float): Angle of the plane in degrees.

    Returns:
        pd.DataFrame: The modified dataframe with corrected wind direction data.
    """
    df = df.copy()
    df["winddir_rel"] = df.apply(lambda row: abs(90 - min_angular_displacement(row["winddir"], plane_angle)), axis=1)  # type: ignore
    df["windspeed_measured"] = df["windspeed"]
    df["windspeed"] = df["windspeed"] * np.cos(np.radians(df["winddir_rel"]))
    return df


def bimodal_azimuth(
    df: pd.DataFrame, course_col: str = "course_azimuth", min_height: int = 5, min_diff: int = 160
) -> tuple[float, float]:
    """
    Identifies the two most frequent course azimuths in the dataset, ensuring they are sufficiently
    distinct. Filters data by height and removes NaNs before analysis.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        course_col (str): Column name for course azimuth data. Default is "course_azimuth".
        min_altitude (int): Minimum height (_ato typically) for data to be included. Default is 5.
        min_diff (int): Minimum difference between the two modes. Default is 160.

    Returns:
        tuple: Two modes of the course azimuth.
    """
    df = df[df["height_ato"] >= min_height]
    data = df[course_col].dropna().to_numpy()
    hist, edges = np.histogram(data, bins=50)
    edgedist = edges[1] - edges[0]
    bin_centers = edges[:-1] + edgedist / 2
    max_freq_idx = np.argsort(hist)[-2:]  #  top 2 frequencies
    mode1, mode2 = bin_centers[max_freq_idx]
    while np.abs(mode1 - mode2) < min_diff:
        if hist[max_freq_idx[0]] < hist[max_freq_idx[1]]:
            hist[max_freq_idx[0]] = 0
            max_freq_idx = np.argsort(hist)[-2:]
        else:
            hist[max_freq_idx[1]] = 0
            max_freq_idx = np.argsort(hist)[-2:]
        mode1, mode2 = bin_centers[max_freq_idx]
    if min_angular_displacement(mode1, mode2) < 160:
        warnings.warn(
            f"Two modes are close together - this probably should never happen: {mode1:.2f} and {mode2:.2f}",
            UserWarning,
            stacklevel=2,
        )
    return (mode1, mode2)


# this returns modes of slope from -90 to 90 degrees.
def bimodal_elevation(
    df: pd.DataFrame, course_col: str = "course_elevation", min_height: float = 5, max_slope: float = 70
) -> tuple[float, float]:
    """
    Identifies the most frequent course elevation in the dataset, adjusted for vertical movements.
    Filters data by height and removes NaNs before analysis.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        course_col (str): Column name for course elevation data. Default is "course_elevation".
        min_height (int): Minimum height for data to be included. Default is 5.
        max_slope (int): Maximum slope to consider, avoTupleiding vertical movements. Default is 70.

    Returns:
        tuple: Mode of course elevation and its negative, representing possible ascent/descent angles.
    """
    df = df[df["height_ato"] >= df["height_ato"].min() + min_height]
    data = df[course_col].to_numpy()
    data = np.abs(data[~np.isnan(data)])
    # to get around the edge case where vertical movements are modal
    data = data[data < max_slope]
    hist, edges = np.histogram(data, bins=50)
    max_freq_idx = np.argsort(hist)[::-1][:2]
    mode = edges[max_freq_idx][0]
    return (mode, -mode)


def height_transect_splitter(df: pd.DataFrame, height_col: str = "height_ato") -> tuple[pd.DataFrame, Figure]:
    """
    Splits the dataset into height-based transects and plots histogram peaks to identify prominent
    height ranges. Only works if the flights are flat.

    Parameters:
        df (pd.DataFrame): The input dataframe containing height data.

    Returns:
        tuple: Modified dataframe with transect labels and a figure showing the histogram with peaks.
    """
    df = df.copy()
    heights = df[height_col].to_numpy()
    counts, bin_edges = np.histogram(heights, bins=40)
    bin_width = bin_edges[1] - bin_edges[0]
    bin_edges = np.append(heights.min() - bin_width, bin_edges)  # avoid literal edge effects
    bin_edges = np.append(bin_edges, heights.max() + bin_width)
    counts = np.append(0, counts)
    counts = np.append(counts, 0)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    peaks, properties = find_peaks(counts)
    transect_edges = (bin_centers[peaks][:-1] + bin_centers[peaks][1:]) / 2
    transect_edges = np.append(heights.min(), transect_edges)
    transect_edges = np.append(transect_edges, heights.max())
    fig, ax = plt.subplots()
    ax.stairs(edges=bin_edges, values=counts, fill=True)
    ax.plot(bin_centers[peaks], counts[peaks], "x", color="red")
    ax.vlines(transect_edges, ymin=0, ymax=max(counts), color="red")
    df["transect_num"] = pd.cut(df[height_col], bins=list(transect_edges), labels=False, include_lowest=True)  # type: ignore
    return df, fig


def add_transect_azimuth_switches(df: pd.DataFrame, threshold=150, shift=3) -> pd.DataFrame:
    """
    Identifies transects based on significant changes in course azimuth, incrementing a transect
    counter to distinguish different flight paths. This is a really crude function and should probably
    be only used with data that's already been filtered in some way.

    Parameters:
        df (pd.DataFrame): The input dataframe with course azimuths.

    Returns:
        pd.DataFrame: The modified dataframe with a new 'transect' column indicating transect IDs.
    """
    df = df.copy()
    df["transect_num"] = 0
    df["prev_course_azimuth"] = df["course_azimuth"].shift(shift)  # this gives better behaviour for very neat transects
    df["deg_displace"] = df.apply(
        lambda row: min_angular_displacement(row["course_azimuth"], row["prev_course_azimuth"])  # type: ignore
        if not pd.isnull(row["prev_course_azimuth"])
        else np.nan,
        axis=1,
    )
    df.loc[df["deg_displace"] > threshold, "transect_num"] = 1
    # remove ones that are next to each other
    for i in range(1, len(df)):
        if (df.loc[i, "transect_num"] == 1 and df.loc[i - 1, "transect_num"] == 1) or (
            df.loc[i, "transect_num"] == 1 and df.loc[i - 2, "transect_num"] == 1
        ):
            df.loc[i, "transect_num"] = 0
    df["transect_num"] = df["transect_num"].shift(-shift)  # recorrect for shift
    df["transect_num"] = df["transect_num"].cumsum() + 1  # 1 indexed
    df["transect_num"] = df["transect_num"].ffill()
    df = df.drop(columns=["prev_course_azimuth", "deg_displace"])
    return df


def course_filter(
    df: pd.DataFrame, azimuth_filter: float, azimuth_window: int, elevation_filter: float
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Filters data based on specified course azimuth and elevation, aiming to isolate transects that
    align with main flight directions. First elevation is filtered to remove significant climbs or descents
    (beware terrain-following flights). Bimodal course azimuths are calculated and used to filter the
    data based on the main flight directions, with a rolling median applied (the window) to smooth the data.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        azimuth_filter (float): The tolerance for deviation from the main course azimuths.
        azimuth_window (int): The window size for rolling median calculation of course azimuths.
        elevation_filter (float): The tolerance for deviation from horizontal flight.

    Returns:
        tuple: The filtered dataframe and the original unfiltered dataframe for comparison.
    """
    df_filtered = df.copy()
    df_filtered = df_filtered[abs(df["course_elevation"]) < elevation_filter]

    azi1, azi2 = bimodal_azimuth(df_filtered)
    logging.info(f"Drone appears to be flying mainly on the courses {azi1:.2f} degrees and {azi2:.2f} degrees")

    df_filtered["rolling_course_azimuth"] = (
        df_filtered["course_azimuth"].rolling(azimuth_window, center=True).apply(lambda x: circ_median(x), raw=True)
    )

    df_filtered = df_filtered[
        (df_filtered["rolling_course_azimuth"] < azi1 + azimuth_filter)
        & (df_filtered["rolling_course_azimuth"] > azi1 - azimuth_filter)
        | (df_filtered["rolling_course_azimuth"] < azi2 + azimuth_filter)
        & (df_filtered["rolling_course_azimuth"] > azi2 - azimuth_filter)
    ]

    return df_filtered, df


def mCount_max(data_dict: dict[int, float]) -> tuple[int, int]:
    """
    Finds the start and end of the longest monotonic sequence in a dictionary of floats, typically used
    to identify a series of continuous altitude measurements. The first and last are retained.

    Parameters:
    data_dict (Dict[int, float]): Dictionary with sequential numeric keys and numeric values representing measures
    such as altitude.

    Returns:
    tuple: Start and end indices of the longest monotonic sequence in the dictionary.
    """
    if len(data_dict) < 2:
        raise ValueError("Dictionary must contain at least two values")
    if list(data_dict.keys()) != list(range(1, len(data_dict) + 1)):
        raise ValueError("Keys must be sequential integers starting from 1")
    poscount = 0
    negcount = 0
    max_pos_count = 0
    max_pos_transect = 0
    max_neg_count = 0
    max_neg_transect = 0
    pos_start = 0
    neg_start = 0

    for i in range(2, len(data_dict) + 1):
        if data_dict[i] >= data_dict[i - 1]:
            poscount += 1
            negcount = 0
        elif data_dict[i] < data_dict[i - 1]:
            negcount += 1
            poscount = 0

        if max_pos_count < poscount:
            max_pos_count = poscount
            max_pos_transect = i
            pos_start = i - poscount
        elif max_neg_count < negcount:
            max_neg_count = negcount
            max_neg_transect = i
            neg_start = i - negcount

    if max_pos_count > 0 or max_neg_count > 0:
        if max_pos_count >= max_neg_count:
            return pos_start, max_pos_transect
        else:
            return neg_start, max_neg_transect
    else:
        return 0, 0


def largest_monotonic_transect_series(
    df: pd.DataFrame, transect_col: str = "transect_num", alt_col: str = "height_ato"
) -> tuple[pd.DataFrame, int, int]:
    """
    Filters the input dataframe to include only the largest continuous series of transects based on
    monotonic altitude changes.

    Parameters:
        df (pd.DataFrame): The input dataframe with transect and altitude information.
        transect_col (str): Column name for transect numbers. Default is "transect_num".
        alt_col (str): Column name for altitude data. Default is "height_ato".

    Returns:
        tuple: The filtered dataframe, start transect, and end transect of the largest monotonic series.
    """
    df = add_transect_azimuth_switches(df)  # course switches
    alt_dict = dict(df.groupby(transect_col)[alt_col].mean())
    starttransect, endtransect = mCount_max(alt_dict)  # type: ignore
    # filter to the biggest monotonic series of values
    df = df[(df[transect_col] >= starttransect) & (df[transect_col] <= endtransect)]
    logging.info(
        f"Parsed a flight of {len(np.unique(df[transect_col]))} transects from {alt_dict[starttransect]:.0f}m"
        f" to {alt_dict[endtransect]:.0f}m between {df['timestamp'].iloc[0]} and {df['timestamp'].iloc[-1]}",
    )
    return df, starttransect, endtransect


def monotonic_transect_groups(
    df: pd.DataFrame, transect_col: str = "transect_num", alt_col: str = "height_ato"
) -> tuple[pd.DataFrame, dict[int, str]]:
    """
    Groups transects into a dict of monotonic transect sequences based on altitude, facilitating analysis of continuous
    flight patterns. Current behaviour is to reuse end transects of previous sequences as the start of the next.

    Parameters:
        df (pd.DataFrame): The input dataframe with transect and altitude information.
        transect_col (str): Column name for transect numbers. Default is "transect_num".
        alt_col (str): Column name for altitude data. Default is "height_ato".

    Returns:
        tuple: The dataframe with a new 'group' column indicating the monotonic group ID, and a
        dictionary mapping transects to group IDs.
    """

    df = add_transect_azimuth_switches(df)
    alt_dict = dict(df.groupby(transect_col)[alt_col].mean())

    group_dict = {}
    previous_altitude = None
    current_group = 1
    previous_trend = None
    first_transect_in_series = True

    for transect, altitude in alt_dict.items():
        if previous_altitude is None:
            group_dict[transect] = f"Group_{current_group}"
        else:
            if altitude == previous_altitude:
                raise ValueError("Error: altitude is the same as the previous transect")
            elif altitude > previous_altitude:
                current_trend = "ascending"
            else:
                current_trend = "descending"

            if current_trend != previous_trend and previous_trend is not None and not first_transect_in_series:
                current_group += 1  # Increment group counter
                first_transect_in_series = True
            if current_trend == previous_trend or previous_trend is None:
                # handles edge case where someone flies up, flies down one transect and then up again (yes really)
                first_transect_in_series = False
            group_dict[transect] = f"Group_{current_group}"
            previous_trend = current_trend
        previous_altitude = altitude
    df["group"] = df["transect_num"].map(group_dict)

    return df, group_dict


def remove_non_transects(
    df: pd.DataFrame,
    chain_length: int = 70,
    azimuth_tolerance: int = 10,
    elevation_tolerance: int = 40,
    smoothing_window: int = 5,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Filters the dataframe to remove segments not matching the criteria for being considered as transects,
    based on course azimuth and elevation and segment length.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        chain_length (int): Minimum number of consecutive points to be considered a transect. Default is 70.
        azimuth_tolerance (int): Tolerance for deviation from modal course azimuths. Default is 10.
        elevation_tolerance (int): Tolerance for deviation from modal course elevation. Default is 40.
        smoothing_window (int): Window size for rolling median smoothing of courses. Default is 5. Rolling median
            avoids single point errors.

    Returns:
        tuple: Dataframes of removed segments and retained segments that fit transect criteria.
    """

    def get_true_runs(mask):
        enumerated_mask = list(enumerate(mask))  # add index numbers to mask
        # group consecutive true/false values
        groups = groupby(enumerated_mask, key=lambda x: x[1])
        # retain groups of True values
        true_runs = [list(group) for key, group in groups if key]
        return true_runs

    def split_runs_on_azimuth_inversion(df, runs, azimuth_inversion_threshold=120):
        split_runs = []
        for run in runs:
            last_azimuth = df.iloc[run[0][0]]["smoothed_course_azimuth"]
            current_run = [run[0]]
            for point in run[1:]:
                current_azimuth = df.iloc[point[0]]["smoothed_course_azimuth"]
                if min_angular_displacement(current_azimuth, last_azimuth) > azimuth_inversion_threshold:
                    split_runs.append(current_run)
                    current_run = [point]
                else:
                    current_run.append(point)
                last_azimuth = current_azimuth
            split_runs.append(current_run)
        return split_runs

    # Create new columns for filtering reasons, initialized with False
    df_removed = df.copy()
    df_removed["filtered_by_azimuth"] = False
    df_removed["filtered_by_elevation"] = False
    df_removed["filtered_by_chain"] = False

    # Apply azimuth filter
    modal_course_azimuths = bimodal_azimuth(df, course_col="course_azimuth")
    modal_elevation_azimuths = bimodal_elevation(df, course_col="course_elevation")

    # Apply rolling median to course azimuth and elevation and store them in new columns
    df_removed["smoothed_course_azimuth"] = (
        df_removed["course_azimuth"]
        .rolling(smoothing_window, center=True)
        .apply(lambda x: circmean(x, 360, 0), raw=True)
        .fillna(df_removed["course_azimuth"], inplace=False)
    )

    df_removed["smoothed_course_elevation"] = (
        df_removed["course_elevation"]
        .rolling(smoothing_window, center=True)
        .apply(lambda x: circmean(x, 360, 0), raw=True)
        .fillna(df_removed["course_elevation"], inplace=False)
    )

    azimuth_mask = df_removed["smoothed_course_azimuth"].apply(
        lambda x: any([min_angular_displacement(x, mode) <= azimuth_tolerance for mode in modal_course_azimuths])
    )

    df_removed.loc[~azimuth_mask, "filtered_by_azimuth"] = True

    # Apply elevation filter
    elevation_mask = df_removed["smoothed_course_elevation"].apply(
        lambda x: any([min_angular_displacement(x, mode) <= elevation_tolerance for mode in modal_elevation_azimuths])
    )
    df_removed.loc[~elevation_mask, "filtered_by_elevation"] = True

    # Combined mask for azimuth and elevation
    mask = azimuth_mask & elevation_mask

    true_runs = get_true_runs(mask)
    split_runs = split_runs_on_azimuth_inversion(df_removed, true_runs)
    chain_mask = [len(run) >= chain_length for run in split_runs]
    chain_filtered_runs = [run for run, filter_by_chain in zip(split_runs, chain_mask, strict=False) if filter_by_chain]
    chain_filtered_indices = [point[0] for run in chain_filtered_runs for point in run]
    df_removed.loc[df_removed.index.difference(chain_filtered_indices), "filtered_by_chain"] = True

    filtered_segments = [df_removed.iloc[run[0][0] : run[-1][0] + 1] for run in chain_filtered_runs]

    df_retained = pd.concat(filtered_segments)
    df_removed = df_removed[~df_removed.index.isin(df_retained.index)]

    return df_removed, df_retained


def flatten_linear_plane(
    df: pd.DataFrame, alt_col: str = "height_ato", distance_filter: float = 10000
) -> tuple[pd.DataFrame, float]:
    """
    Transforms a 3D dataset into a linear plane, focusing on the largest contiguous dataset aligned with
    the plane's primary orientation. The new x coordinate is the distance along the plane, y is the distance
    perpendicular to the plane (useful only for deviation), and z is the altitude.

    Parameters:
        df (pd.DataFrame): Input dataframe with "utm_easting", "utm_northing", and "height_ato" columns.
        alt_col (str): Column name for altitude data. Default is "height_ato".
        distance_filter (float): Threshold for filtering points based on their distance from the regression line.

    Returns:
        tuple: Modified dataframe with new 'x', 'y', and 'z' columns representing transformed coordinates,
        and one of the plane's angles of rotation (from N, 0-360) in degrees.
    """

    def orthogonal_distance_regression(df: pd.DataFrame) -> tuple[pd.DataFrame, np.ndarray[Any, Any]]:
        """
        Perform orthogonal distance regression on the given DataFrame.

        Parameters:
            df (pd.DataFrame): DataFrame containing "utm_easting" and "utm_northing" columns.

        Returns:
            tuple[pd.DataFrame, np.ndarra[Any,Any]]]: Updated DataFrame with "distance_from_fit" column
            and the fitted parameters (slope, intercept).
        """

        def linear_reg_equation(B, x):
            return B[0] * x + B[1]  # y = mx + c

        required_columns = ["utm_easting", "utm_northing"]
        if not all(col in df.columns for col in required_columns):
            raise ValueError(f"DataFrame must contain columns: {required_columns}")
        model = odr.Model(linear_reg_equation)
        data = odr.Data(df["utm_easting"], df["utm_northing"])

        INITIAL_BETA = [1, 0]  # Initial guess of slope=1, intercept=0
        odr_instance = odr.ODR(data, model, beta0=INITIAL_BETA)
        fit = odr_instance.run()

        if fit.stopreason[0] == "Iteration limit reached":
            raise RuntimeError("ODR fitting failed to converge")

        slope, intercept = fit.beta
        df = df.assign(
            distance_from_fit=abs((slope * df["utm_easting"] - df["utm_northing"] + intercept) / np.sqrt(slope**2 + 1))
        )

        return df, fit.beta

    df, coefs2D = orthogonal_distance_regression(df)
    df = df.loc[df["distance_from_fit"] < distance_filter, :]
    df, coefs2D = orthogonal_distance_regression(df)  # this is intentionally done twice
    df = df.loc[df["distance_from_fit"] < distance_filter, :]
    rotation = np.arctan(coefs2D[0])
    df.loc[:, "x"] = (df["utm_easting"] - df["utm_easting"].min()) * np.cos(-rotation) - (
        df["utm_northing"] - df["utm_northing"].min()
    ) * np.sin(-rotation)
    df.loc[:, "y"] = (df["utm_easting"] - df["utm_easting"].min()) * np.sin(-rotation) + (
        df["utm_northing"] - df["utm_northing"].min()
    ) * np.cos(-rotation)
    df.loc[:, "z"] = df[alt_col]

    plane_angle = (np.pi / 2) - np.arctan(coefs2D[0])
    plane_angle = np.degrees(plane_angle)
    return df, plane_angle


## Functions for circular/spiral flights ##


def circle_deviation(df: pd.DataFrame, x_col: str, y_col: str) -> tuple[pd.DataFrame, float, float, float]:
    """
    Calculates the deviation of points from a fitted circle and their azimuth angles.

    Parameters:
        df (pd.DataFrame): The input dataframe containing the data.
        x (str): Column name for the x-coordinate.
        y (str): Column name for the y-coordinate.

    Returns:
        Tuple[pd.Dataframe, float, float, float]: A tuple containing the modified dataframe with azimuth angles
        and deviations,the radius of the fitted circle, and the coordinates of the circle's center.
    """
    df = df.copy()
    required_columns = [x_col, y_col]
    if not all(col in df.columns for col in required_columns):
        raise ValueError(f"DataFrame must contain columns: {required_columns}")

    x = df[x_col]
    y = df[y_col]

    def midhalf(x):  # select middle half of data to avoid edge effects
        return x.iloc[int(len(x) * 1 / 4) : int(len(x) * 3 / 4)]

    x_filter, y_filter = midhalf(x), midhalf(y)

    def func(params):
        xc, yc, r = params
        return np.sqrt((x - xc) ** 2 + (y - yc) ** 2) - r

    x_m = np.mean(np.array(x_filter))  # initial guess for parameters
    y_m = np.mean(np.array(y_filter))
    r_m = np.mean(np.sqrt((np.array(x_filter) - x_m) ** 2 + (np.array(y_filter) - y_m) ** 2))

    params0 = np.array([x_m, y_m, r_m])
    result = least_squares(func, params0)
    xc, yc, r = result.x

    deviation = np.sqrt((x - xc) ** 2 + (y - yc) ** 2) - r

    # output azimuth in radians with 0 at north and increasing clockwise thanks to modulos
    azimuth = np.degrees(np.arctan2(x - xc, y - yc) % (2 * np.pi))

    df["circ_azimuth"] = azimuth
    df["circ_deviation"] = deviation

    return df, r, xc, yc


def recentre_azimuth(df: pd.DataFrame, r: float, x: str = "circ_azimuth", y: str = "ch4_normalised") -> pd.DataFrame:
    """
    Recentres the azimuth angles based on the angle of maximum value and computes the distance along the
    circumference of the circle.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        r (float): The radius of the circle.
        x (str): Column name for azimuth angles. Default is 'circ_azimuth'.
        y (str): Column name for the values used to find the maximum. Default is 'ch4_normalised'.

    Returns:
        pd.DataFrame: The modified dataframe with centered azimuth angles and distances along the circumference.
    """
    df = df.copy()

    def azimuth_of_max(df: pd.DataFrame, x: str = "circ_azimuth", y: str = "ch4_normalised") -> float:
        return df.loc[df[y].idxmax()][x]  # type: ignore

    centre_azimuth = azimuth_of_max(df, x, y)
    df["centred_azimuth"] = df[x] - centre_azimuth
    df.loc[df["centred_azimuth"] > 180, "centred_azimuth"] -= 360
    df.loc[df["centred_azimuth"] < -180, "centred_azimuth"] += 360
    df["circumference_distance"] = r * np.radians(df["centred_azimuth"])
    df["circumference_distance"] = df["circumference_distance"] - df["circumference_distance"].min()
    return df


def drone_anemo_to_point_wind(
    df: pd.DataFrame, yaw_col: str, anemo_u_col: str, anemo_v_col: str, easting_col: str, northing_col: str
) -> pd.DataFrame:
    """
    Convert anemometer wind data from drone's coordinate system to Earth's coordinate system
    and calculate true wind speed and direction.

    Parameters:
    df (pd.DataFrame): Input DataFrame containing drone yaw, anemometer data, and drone speed.
    yaw_col (str): Column name for drone's yaw (in degrees, range [-180, 180]).
    anemo_u_col (str): Column name for anemometer U (wind speed in drone's X direction, from port to starboard).
    anemo_v_col (str): Column name for anemometer V (wind speed in drone's Y direction, from aft to nose).
    easting_col (str): Column name for drone's speed from west to east
    northing_col (str): Column name for drone's speed from south to north

    Returns:
    pd.DataFrame: DataFrame with calculated true wind speed ("windspeed") and true wind direction ("winddir").
    """
    yaw_rad = np.deg2rad(df[yaw_col] % 360)
    rotated_U = df[anemo_u_col] * np.cos(yaw_rad) + df[anemo_v_col] * np.sin(yaw_rad)
    rotated_V = -df[anemo_u_col] * np.sin(yaw_rad) + df[anemo_v_col] * np.cos(yaw_rad)
    true_U = -rotated_U - df[easting_col]
    true_V = -rotated_V - df[northing_col]
    df["windspeed"] = np.sqrt(true_U**2 + true_V**2)
    df["winddir"] = np.degrees(np.arctan2(true_U, true_V)) % 360

    return df



"""Baselining functions."""

import numpy as np
import pandas as pd
import pybaselines as pybs
from skimage.filters import threshold_triangle as custom_threshold
from . import plotting


def algorithmic_baseline(
    df: pd.DataFrame,
    gas: str,
    algorithmic_baseline_settings: dict,
):
    df = df.copy()
    algorithm = algorithmic_baseline_settings["algorithm"]
    settings = algorithmic_baseline_settings.get(algorithm, {}).copy()
    if settings.get("threshold") == "custom":
        settings["threshold"] = custom_threshold
    if len(df) < 20:
        raise ValueError("Dataframe must contain at least 20 rows for background correction.")
    index = np.arange(len(df))
    baseline_fitter = pybs.Baseline(index, check_finite=False)
    fit = getattr(baseline_fitter, algorithm)
    bkg, params = fit(df[gas], **settings)
    bkg_points = params["mask"]
    df[f"{gas}_normalised"] = df[gas] - bkg
    df[f"{gas}_fit"] = bkg
    background = (df[gas] - bkg)[bkg_points]
    signal = (df[gas] - bkg)[~bkg_points]
    df[f"{gas}_signal"] = np.invert(bkg_points)
    fig = plotting.background_plotting(df, gas)
    output_text = (
        f"Baseline algorithm: {algorithm}\n"
        f"Positive and negative 95% percentile of baseline: {np.percentile(background, 2.5):.2f} ppm, \
            {np.percentile(background, 97.5):.2f} ppm\n"
        f"Mean of baseline: {np.mean(background):.2f} ppm\n"
        f"Minimum and maximum of baseline: {np.min(background):.2f} ppm, {np.max(background):.2f} ppm\n"
        f"Signal points: {len(signal)}; background points: {len(background)}\n"
    )
    return df, fig, output_text
